/**
\file mathexpr.cpp
\date 25/05/2021
\author Yann Boucher
\version 1
\brief MathExpr

Ce fichier contient une fonction permettant d'exécuter une expression mathématique à partir d'une chaîne de caractères et d'une liste de variables.
        **/

// Basé sur l'algorithme du Shunting Yard, suivi d'une évaluation de l'expression sous notation polonaise inversée résultante.

#include "mathexpr.hpp"

#include <stack>
#include <vector>
#include <algorithm>
#include <cctype>

struct rpl_token_t
{
    enum
    {
        LITERAL,
        OP
    } type;
    union
    {
        int value;
        char op;
    };
};

//! Priorité de l'opérateur
static int precedence(char op)
{
    if (op == '*' || op == '/' || op == '%')
        return 3;
    else if (op == '+' || op == '-')
        return 2;
    else
        return 1;
}

//! Evalue une opération binaire entre x et y, dénotée par op.
static int eval_int_binop(char op, int x, int y)
{
    if ((op == '/' || op == '%') && y == 0)
        throw MathExprException("Divsion by zero error");

    switch (op)
    {
        case '+':
            return x+y;
        case '-':
            return x-y;
        case '*':
            return x*y;
        case '/':
            return x/y;
        case '%':
            return x%y;
        default:
            throw MathExprException("Unknown operator");
    }
}

//! Evalue une expression RPL
static int evaluate_rpl_input(const std::vector<rpl_token_t>& rpl_stack)
{
    std::vector<int> data_stack;

    for (unsigned i = 0; i < rpl_stack.size(); ++i)
    {
        if (rpl_stack[i].type == rpl_token_t::LITERAL)
            data_stack.push_back(rpl_stack[i].value);
        else
        {
            int op = rpl_stack[i].op;
            if (data_stack.size() <= 1)
                throw MathExprException("Invalid math expression error");

            int x = data_stack[data_stack.size()-2];
            int y = data_stack[data_stack.size()-1];
            int result = eval_int_binop(op, x, y);
            data_stack.pop_back();
            data_stack[data_stack.size()-1] = result;

        }
    }

    if (data_stack.size() != 1)
        throw MathExprException("Invalid math expression error");

    return data_stack[0];
}

int eval_math(const std::string &in_expr, const std::map<std::string, int> &variables)
{
    std::string expr = in_expr;
    // remove whitespace out of the equation altogether
    expr.erase(std::remove_if(expr.begin(), expr.end(), ::isspace), expr.end());

    std::vector<rpl_token_t> rpl_stack;
    std::stack<char> op_stack;

    rpl_token_t rpl_token;
    rpl_token.type = rpl_token_t::LITERAL; rpl_token.value = 0; // silence warnings

    char* end_ptr;
    unsigned idx = 0;
    while (idx < expr.size())
    {
        switch (expr[idx])
        {
            case '0'...'9':
            {
                int value = std::strtol(&expr.c_str()[idx], &end_ptr, 10);
                int read_len = end_ptr - &expr.c_str()[idx];
                idx += read_len;

                rpl_token.type = rpl_token_t::LITERAL;
                rpl_token.value = value;
                rpl_stack.push_back(rpl_token);
                break;
            }
            case '-':
            case '+':
            case '*':
            case '/':
            case '%':
                // binary operator
            {
                while  (!op_stack.empty() &&
                        (op_stack.top() != '(') &&
                        (precedence(op_stack.top()) > precedence(expr[idx])))
                {
                    rpl_token.type = rpl_token_t::OP;
                    rpl_token.op = op_stack.top();
                    rpl_stack.push_back(rpl_token);
                    op_stack.pop();
                }
                op_stack.push(expr[idx]);
                ++idx;
            }
                break;
            case '(':
                op_stack.push('(');
                ++idx;
                break;
            case ')':
                while (!op_stack.empty() &&
                       op_stack.top() != '(')
                {
                    rpl_token.type = rpl_token_t::OP;
                    rpl_token.op = op_stack.top();
                    rpl_stack.push_back(rpl_token);
                    op_stack.pop();
                }

                if (op_stack.empty())
                    throw MathExprException("Parenthesis mismatch");
                op_stack.pop();

                ++idx;
                break;
            default:
                // assume it's a variable name then
            {
                std::string var_name;
                while (isalnum(expr[idx]))
                {
                    var_name += expr[idx];
                    ++idx;
                }

                if (!variables.count(var_name))
                    throw MathExprException("No such variable '" + var_name + "' found");
                rpl_token.type = rpl_token_t::LITERAL;
                rpl_token.value = variables.at(var_name);
                rpl_stack.push_back(rpl_token);
                break;
            }
        }
    }
    // while there are remaining operators on the op stack
    while (!op_stack.empty())
    {
        rpl_token.type = rpl_token_t::OP;
        rpl_token.op   = op_stack.top();
        op_stack.pop();
        rpl_stack.push_back(rpl_token);
    }


    return evaluate_rpl_input(rpl_stack);
}
